(ns habits.index
  (:require ["ink" :refer [render Box Text Spacer useInput useApp]]
            ["csv-parse/sync" :as csv]
            ["fs" :as fs]
            [habits.components.table :as table]
            [reagent.core :as r]))

; Things This Project would benefit from
; Splitting this single file into smaller files
; Testing
; Types + Type checking
; A working LSP

; Each habit is just a map with keys :checkmarks and :name
; We'll probably need :color and maybe a couple other things, order?
; :checkmarks is an array of maps with keys :date :result
#_(def habits-type [{:name str
                     :checkmarks [{:date str
                                   :result int}]}])

; Config section
(defonce size (r/atom [5 3]))
(defonce scroll-index (r/atom (map dec @size))) ; Position within viewable frame defined by size
(defonce select-index (r/atom [0 0])) ; Position within all the data loaded
(defonce uhabits-csv (r/atom "./data/Loop Habits CSV 2023-05-30/Checkmarks.csv"))

; Mock Data
(def checkmarks_data
  [{:date "27/09/12" :result -1}
   {:date "26/09/12" :result 2}
   {:date "25/09/12" :result 1}
   {:date "24/09/12" :result 0}
   {:date "23/09/12" :result 0}])

(def meditate
  {:name "meditate"
   :checkmarks checkmarks_data})

(def exercise
  {:name "exercise"
   :checkmarks checkmarks_data})

; CSV Parsing
(defn uhabits-data [csv-file]
  (js->clj (csv/parse (str (fs/readFileSync csv-file)) {:columns true})))

(defn parse-row [habits row]
  (let [date (first row)
        values (drop-last (rest row))
        checkmarks (->> values
                        (mapv (fn [v] {:result (js/parseInt v) :date date})))]
    (for [[index habit] (map-indexed vector habits)]
      (assoc habit :checkmarks (conj (:checkmarks habit) (nth checkmarks index))))))

(defn make-habits [csv]
  (let [headers (drop-last (drop 1 (first csv)))
        data (take 20 (rest csv))
        habits (mapv (fn [x] {:name x :checkmarks []}) headers)]
    (vec (reduce parse-row habits data))))

(def my-data (make-habits (uhabits-data @uhabits-csv)))

; Data manipulation
(defonce habits (r/atom my-data))

(defn get-dates [habits]
  (let [dates (->> habits
                   (first)
                   (:checkmarks)
                   (mapv #(get % :date)))]
    dates))

(defn toggle-habits [habits]
  (let [selected @select-index
        scrolled @scroll-index
        pos (->> @size
                 (map #(- % 1))
                 (map - scrolled))
        [i j] (map + pos selected)
        habit (nth habits i)
        checkmarks (:checkmarks habit)
        checkmark (nth checkmarks j)
        new-checkmark (if (<= (:result checkmark) 0) 1 0)
        new-checkmarks (assoc checkmarks j {:date (:date checkmark) :result new-checkmark})
        new-habits (assoc habits i {:name (:name habit)
                                    :checkmarks new-checkmarks})]
    new-habits))

(defn filter-checkmarks [checkmarks]
  (let [[_ date-size] @size
        [_ y] @scroll-index
        y-adj (if (>= y date-size) (- y (- date-size 1)) 0)]
    (->> checkmarks
         (drop y-adj)
         (take date-size))))

(defn filter-habits [habits]
  (let [[habit-size _] @size
        [x _] @scroll-index
        x-adj (if (>= x habit-size) (- x (- habit-size 1)) 0)
        filtered-habits (->> habits
                             (drop x-adj)
                             (take habit-size)
                             (map #(assoc % :checkmarks (filter-checkmarks (:checkmarks %)))))]
    filtered-habits))

(defn update-index [dir]
  (let [[habit-size date-size] @size
        habit-limit (- habit-size 1)
        date-limit (- date-size 1)
        [cur-x cur-y] @select-index]
    (case dir
      "down" (doall (swap! scroll-index (fn [[x y]] (if (>= cur-x habit-limit) [(inc x) y] [x y])))
                    (swap! select-index (fn [[x y]] (if (>= x habit-limit) [habit-limit y] [(inc x) y]))))
      "up" (doall (swap! scroll-index (fn [[x y]] (if (= cur-x 0)
                                                    (if (<= x habit-limit)
                                                      [habit-limit y]
                                                      [(dec x) y])
                                                    [x y])))
                  (swap! select-index (fn [[x y]] (if (<= x 0) [0 y] [(dec x) y]))))
      "right" (doall
               (swap! scroll-index (fn [[x y]] (if (>= cur-y date-limit) [x (inc y)] [x y])))
               (swap! select-index (fn [[x y]] (if (>= y date-limit) [x date-limit] [x (inc y)]))))
      "left" (doall (swap! scroll-index (fn [[x y]] (if (= cur-y 0)
                                                      (if (<= y date-limit)
                                                        [x date-limit]
                                                        [x (dec y)])
                                                      [x y])))
                    (swap! select-index (fn [[x y]] (if (<= y 0) [x 0] [x (dec y)])))))))

; UI
(defn headers [dates]
  [:> Box {:width "100%" :alignItems "center"}
   [:> Spacer]
   (for [date dates]
     [:<> {:key (str "header:" date)}
      [:> Box {:minWidth 10}
       [:> Spacer]]
      [:> Text date]
      [:> Spacer]])])

(defn check-box [isChecked isSelected]
  (let [mark (if (>= isChecked 1) "✔"  "✗")]
    (if isSelected
      [:> Text {:color "yellow"} mark]
      [:> Text mark])))

(defn row [checkmarks habit j]
  [:> Box {:borderStyle "round" :width "100%" :paddingRight 2 :alignItems "center" :flexGrow 1}
   (for [[index checkmark] (map-indexed vector checkmarks)]
     (let [isSelected (= j index)]
       [:<> {:key (str "row:" habit (:date checkmark))}
        [:> Spacer]
        (check-box (:result checkmark) isSelected)
        [:> Spacer]]))])

(defn column [habits i j]
  [:> Box {:borderStyle "round" :flexDirection "column" :columnGap 1}
   (headers (get-dates habits))
   (for [[index habit] (map-indexed vector habits)]
     (let [checkmarks (:checkmarks habit)
           n (:name habit)]
       [:> Box {:key (str "box:row:" n) :alignItems "flex-end"}
        (if (= index i)
          [:<> {:key (str "row:" n)}
           [:> Box {:minWidth 10 :alignSelf "center"}
            [:> Text {:color "yellow"} n]]
           (row checkmarks n j)]
          [:<> {:key (str "row:" n)}
           [:> Box {:minWidth 10 :alignSelf "center"}
            [:> Text n]]
           (row checkmarks n false)])
        [:> Spacer]]))])

(defn selector []
  (let [[i j] @select-index
        habits_ (filter-habits @habits)
        exit (useApp)]
    (useInput (fn [input k]
                (if (= input "q")
                  (. exit exit))
                (if (= input " ")
                  (swap! habits toggle-habits))
                (if k.downArrow
                  (update-index "down"))
                (if k.upArrow
                  (update-index "up"))
                (if k.rightArrow
                  (update-index "right"))
                (if k.leftArrow
                  (update-index "left"))))
    (column habits_ i j)))

(render (r/as-element [:f> selector]))
